package ctbrec.recorder.download.hls;

import static ctbrec.recorder.download.StreamSource.*;

import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.time.Instant;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.xml.bind.JAXBException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.iheartradio.m3u8.ParseException;
import com.iheartradio.m3u8.PlaylistException;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.OS;
import ctbrec.Recording;
import ctbrec.recorder.download.AbstractDownload;
import ctbrec.recorder.download.StreamSource;

public class HlsdlDownload extends AbstractDownload {

    private static final transient Logger LOG = LoggerFactory.getLogger(HlsdlDownload.class);

    protected Model model;
    protected File targetFile;

    protected transient Config config;
    protected transient Process hlsdlProcess;
    protected transient boolean running;

    @Override
    public void init(Config config, Model model, Instant startTime) {
        super.startTime = startTime;
        this.config = config;
        this.model = model;
        String fileSuffix = config.getSettings().ffmpegFileSuffix;
        targetFile = config.getFileForRecording(model, fileSuffix, startTime);
        splittingStrategy = initSplittingStrategy(config.getSettings());
    }

    @Override
    public void start() throws IOException {
        try {
            running = true;
            String threadName = "Download " + model.getName();
            Thread.currentThread().setName(threadName);
            Files.createDirectories(targetFile.getParentFile().toPath());

            Thread splittingMonitor = new Thread(() -> {
                try {
                    while (running) {
                        if (splittingStrategy.splitNecessary(HlsdlDownload.this)) {
                            LOG.debug("splitting strategy {} triggered split", splittingStrategy.getClass().getSimpleName());
                            stop();
                        }
                        TimeUnit.SECONDS.sleep(1);
                    }
                } catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            });
            splittingMonitor.setName(threadName + " splitter");
            splittingMonitor.setDaemon(true);
            splittingMonitor.start();

            Hlsdl hlsdl = new Hlsdl.Builder()
                    .logOutput(config.getSettings().loghlsdlOutput)
                    .onStarted(p -> hlsdlProcess = p)
                    .build();
            String[] cmdline = createCommandLine();
            hlsdl.exec(cmdline, OS.getEnvironment(), targetFile.getParentFile());
        } catch (ParseException e) {
            throw new IOException("Couldn't parse stream information", e);
        } catch (PlaylistException e) {
            throw new IOException("Couldn't parse HLS playlist", e);
        } catch (EOFException e) {
            // end of playlist reached
            LOG.debug("Reached end of playlist for model {}", model);
        } catch (Exception e) {
            throw new IOException("Exception while downloading segments", e);
        } finally {
            stop();
            running = false;
            LOG.debug("Download for {} terminated", model);
        }
    }

    private String[] createCommandLine() throws IOException, ExecutionException, ParseException, PlaylistException, JAXBException {
        String playlistUrl = getSegmentPlaylistUrl(model);
        Map<String, String> headers = model.getHttpHeaderFactory().createSegmentPlaylistHeaders();
        String[] cmdline = new String[9 + headers.size() * 2];
        int idx = 0;
        cmdline[idx++] = config.getSettings().hlsdlExecutable;
        cmdline[idx++] = "-c";
        cmdline[idx++] = "-r";
        cmdline[idx++] = "3";
        cmdline[idx++] = "-w";
        cmdline[idx++] = "3";
        for (Entry<String, String> header : headers.entrySet()) {
            cmdline[idx++] = "-h";
            cmdline[idx++] = header.getKey() + ": " + header.getValue();
        }
        cmdline[idx++] = "-o";
        cmdline[idx++] = targetFile.getCanonicalPath();
        cmdline[idx] = playlistUrl;
        return cmdline;
    }

    protected String getSegmentPlaylistUrl(Model model) throws IOException, ExecutionException, ParseException, PlaylistException, JAXBException {
        LOG.debug("{} stream idx: {}", model.getName(), model.getStreamUrlIndex());
        List<StreamSource> streamSources = model.getStreamSources();
        Collections.sort(streamSources);
        for (StreamSource streamSource : streamSources) {
            LOG.debug("{} src {}", model.getName(), streamSource);
        }
        String url = null;
        if (model.getStreamUrlIndex() >= 0 && model.getStreamUrlIndex() < streamSources.size()) {
            // TODO don't use the index, but the bandwidth. if the bandwidth does not match, take the closest one
            LOG.debug("{} selected {}", model.getName(), streamSources.get(model.getStreamUrlIndex()));
            url = streamSources.get(model.getStreamUrlIndex()).getMediaPlaylistUrl();
        } else {
            // filter out stream resolutions, which are out of range of the configured min and max
            int minRes = Config.getInstance().getSettings().minimumResolution;
            int maxRes = Config.getInstance().getSettings().maximumResolution;
            List<StreamSource> filteredStreamSources = streamSources.stream()
                    .filter(src -> src.height == 0 || src.height == UNKNOWN || minRes <= src.height)
                    .filter(src -> src.height == 0 || src.height == UNKNOWN || maxRes >= src.height)
                    .collect(Collectors.toList());

            if (filteredStreamSources.isEmpty()) {
                throw new ExecutionException(new RuntimeException("No stream left in playlist"));
            } else {
                LOG.debug("{} selected {}", model.getName(), filteredStreamSources.get(filteredStreamSources.size() - 1));
                url = filteredStreamSources.get(filteredStreamSources.size() - 1).getMediaPlaylistUrl();
            }
        }
        LOG.debug("Segment playlist url {}", url);
        return url;
    }

    @Override
    public void stop() {
        if (running) {
            running = false;
            if (hlsdlProcess != null) {
                hlsdlProcess.destroy();
                if (hlsdlProcess.isAlive()) {
                    LOG.info("hlsdl didn't terminate. Destroying the process with force!");
                    hlsdlProcess.destroyForcibly();
                    hlsdlProcess = null;
                }
            }
        }
    }

    @Override
    public void postprocess(Recording recording) {
        // nothing to do
    }

    @Override
    public File getTarget() {
        return targetFile;
    }

    @Override
    public String getPath(Model model) {
        String absolutePath = targetFile.getAbsolutePath();
        String recordingsDir = Config.getInstance().getSettings().recordingsDir;
        String relativePath = absolutePath.replaceFirst(Pattern.quote(recordingsDir), "");
        return relativePath;
    }

    @Override
    public boolean isSingleFile() {
        return true;
    }

    @Override
    public long getSizeInByte() {
        return getTarget().length();
    }

    @Override
    public Model getModel() {
        return model;
    }
}
